package com.exh.bevel;

import java.util.LinkedList;
import java.util.Vector;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Rect;
import android.util.AttributeSet;
import android.view.MotionEvent;
import android.view.View;

import com.exh.bevel.algo.Clip;
import com.exh.bevel.algo.Mask;
import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Polygon;
import com.exh.math.primitive.Tri;

/* BEVEL-VIEW
 * The display surface for any generated bitmaps.
 * NOTE:  This class is primarily behaviour that ets
 * switched on when EDIT_MODE is true, for testing
 * the algorithm interactively.
 */
public class BevelView extends View {

	public final static boolean		EDIT_MODE = false;
	
	private final static boolean	BYPASS_CLIPPER = false;
	
	private final static int	CFG_EQUAL = 0;
	private final static int	CFG_SMALLER = 1;
	private final static int	CFG_LARGER = 2;
	private final static int	CFG_SILO_NOTCH_A = 3;
	private final static int	CFG_SILO_NOTCH_B = 4;
	private final static int	CFG_SILO_TEST_1 = 5;
	private final static int	CFG_SILO_TEST_2 = 6;
	private final static int	CFG_SILO_TEST_3 = 7;
	private final static int	CFG_SILO_TEST_4 = 8;
	private final static int	CFG_SILO_TEST_5 = 9;
	private final static int	CFG_SILO_TEST_6 = 10;
	private final static int	CFG = CFG_SILO_TEST_6;
	
	private final static boolean TEST_REVERSE_A = false;
	private final static boolean TEST_CONTAINS = false;
	private final static boolean TEST_CORNER = true;

	private Bitmap					mBm;
	
	private final Clip				mClipper = new Clip();
	private final Mask				mTriClip = new Mask();
	
	private final LinkedList<Tri>	mTriList = new LinkedList<Tri>();
	private final LinkedList<Tri>	mOutput = new LinkedList<Tri>();
	
	private final Paint			mBgP = new Paint(Paint.ANTI_ALIAS_FLAG),
								mFgP = new Paint(Paint.ANTI_ALIAS_FLAG),
								mIntersectLineP = new Paint(Paint.ANTI_ALIAS_FLAG),
								mIntersectPlaneP = new Paint(Paint.ANTI_ALIAS_FLAG),
								mIntersectTriP = new Paint(Paint.ANTI_ALIAS_FLAG);
	
	private final Tri			mTri0 = new Tri(),
								mTri1 = new Tri();
	private final Polygon		mPolyA = new Polygon();
	private float				mPos0 = 0.35f, mPos1 = 1.0f,
								mTop = 0, mBottom = 0, mRange = 0;
	private final Rect			mF = new Rect();
	private boolean				mTracking = false;
	
	private boolean				mDirty = true;
	
	private final Point3.Order	mOrder = new Point3.Order();
	
	private final LinkedList<Tri> mTri = new LinkedList<Tri>();
	
	// Points to display
	private final Vector<Point3> mBadPts = new Vector<Point3>();
	private final Vector<Point3> mGoodPts = new Vector<Point3>();
	
	public BevelView(Context context, AttributeSet attrs) {
		super(context, attrs);

		mBgP.setColor(0xff000000);
		mFgP.setColor(0xdd00ff00);
		mIntersectLineP.setColor(0x6600ff00);
		mIntersectPlaneP.setColor(0xccff0000);
		mIntersectTriP.setColor(0xee00ff00);
	}
	
	public void setImageBitmap(Bitmap bm) {
		mBm = bm;
		invalidate();
	}
	
	@Override public void onDraw(Canvas c) {
		super.onDraw(c);
		
		if (mBm != null) {
			final int		l = (mF.width()/2) - (mBm.getWidth()/2),
							t = (mF.height()/2) - (mBm.getHeight()/2);
			c.drawBitmap(mBm, l, t, null);
			drawTest(c);
			return;
		}
		
		if (!EDIT_MODE) return;
		
		if (mDirty) {
			cache();
			mDirty = false;
		}
		
		final float		PT_SIZE = 4;
		onDrawTri(mTri0, c, mBgP);
		onDrawTri(mTri1, c, mBgP);
		
for (Tri t : mOutput) System.out.println("OUT=" + t);
		for (Tri t : mOutput) onDrawTri(t, c, mFgP);
//		for (Point3 pt : mClipper.mDisplayPts0) c.drawCircle((float)pt.x, (float)pt.y, PT_SIZE, mIntersectPlaneP);
//		for (Point3 pt : mClipper.mDisplayPts1) c.drawCircle((float)pt.x, (float)pt.y, PT_SIZE, mIntersectTriP);
		for (Point3 pt : mBadPts) c.drawCircle((float)pt.x, (float)pt.y, PT_SIZE, mIntersectPlaneP);
		for (Point3 pt : mGoodPts) c.drawCircle((float)pt.x, (float)pt.y, PT_SIZE, mIntersectTriP);

		drawTest(c);
	}

	private void drawTest(final Canvas c) {
/*
		final Polygon		dst = new Polygon(), src = new Polygon();
//		dst.add(4.00, 190.00, 62);	dst.add(-24.00, 190.00, 90);	dst.add(4.00, 203.29, 62);
//		src.add(35.00, 130.07, 31);	src.add(4.00, 66.00, 62);		src.add(4.00, 252.00, 62);	src.add(35.00, 252.00, 31);
//		dst.add(129.64, 67.64, 90);		dst.add(94.66, 73.20, 69.20);	dst.add(67.64, 129.64, 90.00);
//		src.add(190.00, 4.00, 87.68);	src.add(94.66, 73.20, 69.20);	src.add(103.28, 94.00, 90);	src.add(190.00, 94.00, 151.32);
//		dst.add(199.85, 56.15, 52.15);	dst.add(162.00, 128.00, 90);	dst.add(252.00, 128.00, 0);
//		src.add(215.29, 91.29, 36.71);	src.add(190.00, 128.00, 62);	src.add(252.00, 128.00, 0);
//		dst.add(100.00, 128.00, 90);	dst.add(100.00, 227.32, 90);	dst.add(137.85, 199.85, 52.15);
//		src.add(100.00, 162.00, 90);	src.add(66.00, 162.00, 124);	src.add(66.00, 252.00, 124);	src.add(137.85, 199.85, 52.15);
		dst.add(4.00, 190.00, 62);		dst.add(-24.00, 190.00, 90);	dst.add(4.00, 203.29, 62);
		src.add(35.00, 130.07, 31);		src.add(4.00, 66.00, 62);		src.add(4.00, 252.00, 62);	src.add(35.00, 252.00, 31);
		final Paint			p = new Paint(Paint.ANTI_ALIAS_FLAG);
		p.setStrokeWidth(1);
		p.setColor(0xffff0000);
		drawPoly(c, dst, p, 40, 0);
		p.setColor(0xff00ff00);
		drawPoly(c, src, p, 40, 0);

		final PolyIsExternal	isExt = new PolyIsExternal();
		System.out.println("external dst=" + isExt.on(dst, src));
*/
	}

	private final void drawPoly(final Canvas c, final Polygon poly, final Paint p,
								final double x, final double y) {
		if (poly.mSize < 2) return;
		Point3		prv = poly.mPt[poly.mSize-1];
		for (int k=0; k<poly.mSize; k++) {
			final Point3	nxt = poly.mPt[k];
			c.drawLine((float)(prv.x+x), (float)(prv.y+y), (float)(nxt.x+x), (float)(nxt.y+y), p);
			prv = nxt;
		}
	}

	@Override protected void onSizeChanged(int iw, int ih, int oldw, int oldh) {
		super.onSizeChanged(iw, ih, oldw, oldh);

		mF.set(0, 0, iw, ih);
		
		mDirty = true;
		mTriList.clear();

		final float			w = ((float)(iw));
		final float			pad = w * 0.1f,
							triW = (w - pad*2) * 0.75f;
		mTop = pad;
		mBottom = ih-pad;
		mRange = (mBottom - mTop) - triW;
		double				aL = pad, aR = pad+triW,
							azL = 0, azR = triW;
		if (TEST_REVERSE_A) {
			aL = pad+triW; aR = pad;
			azL = triW; azR = 0;
			mPos1 = 0.5f;
		}
		
		mTri0.mPt[0].set(aL, pad+triW, azL);
		mTri0.mPt[1].set(aR, pad+(triW/2), azR);
		mTri0.mPt[2].set(aL, pad, azL);

		if (TEST_CORNER) {
			final double	d = (w-pad*2);
			mTri0.mPt[0].set(pad, pad, 0);
			mTri0.mPt[1].set(w*0.5, pad+d, d);
			mTri0.mPt[2].set(w-pad, pad, 0);
			mPos0 = 0;
			mPos1 = 0.09f;
		}
		
		mTri1.mPt[0].set(w-pad, pad, 0);
		mTri1.mPt[1].set(w-pad-triW, pad+(triW/2), triW);
		mTri1.mPt[2].set(w-pad, pad+triW, 0);
		if (TEST_CONTAINS) {
			final double		Z_DELTA = -(triW*0.25);
			for (Point3 pt : mTri1.mPt) pt.z += Z_DELTA;
		}
		
		if (CFG == CFG_SMALLER) {
			mTri1.mPt[2].y -= triW/2;
			mTri1.mPt[1].y -= triW/4;
		} else if (CFG == CFG_LARGER) {
			mTri0.mPt[0].y -= triW/2;
			mTri0.mPt[1].y -= triW/4;
		} else if (CFG == CFG_SILO_NOTCH_A) {
			mTri0.set(4, 66, 0,		94, 252, 90,		94, 66, 90);
			mTri1.set(4, 66, 0,		67.64, 129.64, 90,	66, 4, 0);
		} else if (CFG == CFG_SILO_NOTCH_B) {
			mTri0.set(4, 66, 0,		94, 252, 90,		94, 66, 90);
			mTri1.set(66, 4, 0,		67.64, 129.64, 90,	129.64, 67.64, 90);
		} else if (CFG == CFG_SILO_TEST_1) {
			mTri0.set(4, 252, 0,	44, 252, 40,	4, 66, 0);
			mTri1.set(4, 66, 0,		44, 252, 40,		44, 66, 40);
			mPos0 = 0.5f;			mPos1 = 0.5f;
		} else if (CFG == CFG_SILO_TEST_2) {
			mTri0.set(4, 66, 0,		44, 252, 40,		44, 66, 40);
			mTri1.set(4, 66, 0,		32.28, 94.28, 40,	66, 4, 0);
			mPos0 = -1;				mPos1 = -1;
		} else if (CFG == CFG_SILO_TEST_3) {
			mTri0.set(252, 252, 0,	252, 212, 40,	190, 252, 0);
			mTri1.set(252, 252, 0,	212, 66, 40,	212, 252, 40);
			mPos0 = -1;				mPos1 = -1;
		} else if (CFG == CFG_SILO_TEST_4) {
			mTri0.set(227.69, 227.69, 24.31,	190, 212, 40,	190, 252, 0);
			mTri1.set(252, 252, 0,	212, 66, 40,	212, 252, 40);
			mPos0 = -1;				mPos1 = -1;
		} else if (CFG == CFG_SILO_TEST_5) {
			mTri0.set(190, 190, 0,	230, 252, 40,	230, 190, 40);
			mTri1.set(252, 252, 0,	212, 66, 40,	212, 252, 40);
			mPos0 = -1;				mPos1 = -1;
		} else if (CFG == CFG_SILO_TEST_6) {
//			mTri0.set(246.10, 10.05, 0,				198.69, 57.40, 48.55,		252.00, 252.00, 0);
//			mTri1.set(252.00, 252.00, 0,	202.73, 203.92, 48.08,			4.00, 252.00, 247.93);
			
//			mTri0.set(4, 66, 0,				4, 252, 0,				137, 252, 133);
//			mTri1.set(25.83, 30.46, 133,	-36.17, 104.86, 133,	66, 190, 0);
			
			mTri0.set(66.00, 190.00, 62,	51.62, 132.88, 47.62,		4.00, 252.00, 0);
			mTri1.set(51.62, 132.88, 47.62,		39.06, 167.55, 35.06,	66, 190, 62);

			mPos0 = -1;					mPos1 = -1;
		}

		mOrder.makeCcw(mTri0.mPt);
		mOrder.makeCcw(mTri1.mPt);

		mTriList.add(mTri0);
		mTriList.add(mTri1);
		
		// Special cue to tell it to not change the positioning.
		if (!(mPos0 < 0 && mPos1 < 0)) {
			setPosition(mTri0, mPos0);
			setPosition(mTri1, mPos1);
		}
	}
	
	@Override public boolean onTouchEvent(MotionEvent me) {
		final int			action = me.getAction();
		if (action == MotionEvent.ACTION_DOWN) {
			mTracking = true;
			track(me.getX(), me.getY());
		} else if (action == MotionEvent.ACTION_MOVE) {
			if (mTracking) track(me.getX(), me.getY());
		} else if (action == MotionEvent.ACTION_UP) {
			mTracking = false;
		} else return super.onTouchEvent(me);
		return true;
	}
	
	private final void track(final float x, final float y) {
		float			nxt1 = 0;
		if (y >= mBottom) nxt1 = 1;
		else if (y > mTop) nxt1 = (y-mTop) / (mBottom - mTop);
		if (mPos1 == nxt1) return;

		mPos1 = nxt1;
		mDirty = true;
		invalidate();
	}

	private final void cache() {		
		mTri.clear();
		mGoodPts.clear();
		mBadPts.clear();
		if (!(mPos0 < 0 && mPos1 < 0)) setPosition(mTri1, mPos1);
		
		mOutput.clear();
		if (BYPASS_CLIPPER) {
			mTriClip.start();
			mPolyA.setTo(mTri1);
			mTriClip.addClipPoly(mPolyA);
			mTriClip.clip(mTri0, mOutput);
		} else {
//			mClipper.on(new Point3(0, 0, 1000), mTriList, mOutput, null);
		}
	}

	private final void setPosition(final Tri t, final float pos) {
		final Point3		a = t.mPt[0], b = t.mPt[1], c = t.mPt[2];
		final float			topPrv = Math.min((float)a.y, Math.min((float)b.y, (float)c.y));
		final float			topNxt = mTop + (mRange*pos);
		a.y = topNxt + (a.y-topPrv);
		b.y = topNxt + (b.y-topPrv);
		c.y = topNxt + (c.y-topPrv);
	}

	// DRAWING
	
	private final void onDrawTri(final Tri t, final Canvas c, final Paint p) {
		onDrawPts(t.mPt, t.mPt.length, c, p);
	}
	
	private final void onDrawPts(	final Point3[] pts, final int size,
									final Canvas c, final Paint p) {
		Point3					prv = pts[size-1];
		for (Point3 nxt : pts) {
			c.drawLine((float)prv.x, (float)prv.y, (float)nxt.x, (float)nxt.y, p);
			prv = nxt;
		}
	}
}
