package com.exh.bevel;

import java.util.LinkedList;
import java.util.List;

import android.graphics.RectF;

import com.exh.math.primitive.Geo;
import com.exh.math.primitive.Point3;

/* A collection of hardcoded point lists, one of which
 * is used as the initial input to the algorithm.
 */
public class Input {

	public final List<Point3>		mPts = new LinkedList<Point3>();
	int								mDirection = 0;
	
	public static final Input initSilo() {
		final Input			ans = new Input();
		ans.mPts.add(new Point3(130, 80, 0));
		ans.mPts.add(new Point3(230, 80, 0));
		ans.mPts.add(new Point3(280, 130, 0));
		ans.mPts.add(new Point3(280, 280, 0));
		ans.mPts.add(new Point3(230, 280, 0));
		ans.mPts.add(new Point3(230, 230, 0));
		ans.mPts.add(new Point3(180, 170, 0));
		ans.mPts.add(new Point3(130, 230, 0));
		ans.mPts.add(new Point3(130, 280, 0));
		ans.mPts.add(new Point3(80, 280, 0));
		ans.mPts.add(new Point3(80, 130, 0));
		return ans;
	}

	public static final Input initFlip() {
		final Input			ans = new Input();
		ans.mPts.add(new Point3(0, 0, 0));
		ans.mPts.add(new Point3(200, 0, 0));
		ans.mPts.add(new Point3(200, 200, 0));
		ans.mPts.add(new Point3(100, 200, 0));
		ans.mPts.add(new Point3(60, 190, 0));
		ans.mPts.add(new Point3(50, 170, 0));
		ans.mPts.add(new Point3(40, 130, 0));
		ans.mPts.add(new Point3(0, 100, 0));
		return ans;
	}

	public static final Input initThrowingStar() {
		final Input			ans = new Input();
		final double		w = 200, h = 200;
		final double		midX = w*0.5, midY = h*0.5;
		final double		x0 = 40;
//		final double		y0 = 20, y1 = 80;
		final double		y0 = 0, y1 = 40;
		ans.mPts.add(new Point3(0, 0, 0));
		ans.mPts.add(new Point3(x0, y0, 0));
		ans.mPts.add(new Point3(midX, y1, 0));
		ans.mPts.add(new Point3(w-x0, y0, 0));
		ans.mPts.add(new Point3(w, 0, 0));
		ans.mPts.add(new Point3(w-y0, 0+x0, 0));
		ans.mPts.add(new Point3(w-y1, midY, 0));
		ans.mPts.add(new Point3(w-y0, h-x0, 0));
		ans.mPts.add(new Point3(w, h, 0));
		ans.mPts.add(new Point3(w-x0, h-y0, 0));
		ans.mPts.add(new Point3(midX, h-y1, 0));
		ans.mPts.add(new Point3(x0, h-y0, 0));
		ans.mPts.add(new Point3(0, h, 0));
		ans.mPts.add(new Point3(0+y0, h-x0, 0));
		ans.mPts.add(new Point3(0+y1, midY, 0));
		ans.mPts.add(new Point3(y0, x0, 0));

		return ans;
	}

	public static final Input initSquare() {
		final Input			ans = new Input();
		ans.mPts.add(new Point3(0, 0, 0));
		ans.mPts.add(new Point3(200, 0, 0));
		ans.mPts.add(new Point3(200, 200, 0));
		ans.mPts.add(new Point3(0, 200, 0));
		return ans;
	}

	public static final Input initQuad() {
		final Input			ans = new Input();
		ans.mPts.add(new Point3(80, 75, 0));
		ans.mPts.add(new Point3(280, 80, 0));
		ans.mPts.add(new Point3(285, 280, 0));
		ans.mPts.add(new Point3(75, 280, 0));
		return ans;
	}

	public static final Input initT() {
		final Input			ans = new Input();
		ans.mPts.add(new Point3(80, 80, 0));
		ans.mPts.add(new Point3(280, 80, 0));
		ans.mPts.add(new Point3(280, 180, 0));
		ans.mPts.add(new Point3(230, 180, 0));
		ans.mPts.add(new Point3(230, 280, 0));
		ans.mPts.add(new Point3(130, 280, 0));
		ans.mPts.add(new Point3(130, 180, 0));
		ans.mPts.add(new Point3(80, 180, 0));
		return ans;
	}
	
	public static final Input initTree() {
		final Input			ans = new Input();
		ans.mDirection = -1;
		final Input			inner = initCircle(45, 90, new RectF(50, 50, 150, 150));
		final Point3		first = inner.mPts.get(1), last = inner.mPts.get(inner.mPts.size()-1);
		
		ans.mPts.add(new Point3(0, 200, 0));
		ans.mPts.add(new Point3(first.x, 200, 0));
		for (int k=1; k<inner.mPts.size()-0; k++) {
			ans.mPts.add(inner.mPts.get(k));
		}
		ans.mPts.add(new Point3(last.x, 200, 0));
		ans.mPts.add(new Point3(200, 200, 0));
		ans.mPts.add(new Point3(200, 0, 0));
		ans.mPts.add(new Point3(0, 0, 0));

		return ans;
	}
	
	public static final Input initCircle(final int steps, final double rotate) {
		return initCircle(steps, rotate, new RectF(0, 0, 256, 256));
	}


	public static final Input initCircle(	final int steps, final double rotate,
											final RectF frame) {
		final Input			ans = new Input();
		final Point3		cen = new Point3(frame.centerX(), frame.centerY(), 0);
		final double		dist = cen.x - frame.left;
		final double		scale = (360.0 / (double)(steps));

		for (int k=0; k<steps; k++) {
			final Point3	pt = new Point3(0, 0, 0);
			double			deg = rotate + (k*scale);
			if (deg >= 360) deg -= 360;
			getCircleEdge(cen, 360-deg, dist, pt);
			ans.mPts.add(pt);
		}
		return ans;
	}

	public static final Input initCornerCircle(final int steps, final double size) {
		final Input			ans = new Input();
		final RectF			b = new RectF(0, 0, 256, 256);
		final double		dist = Math.min(b.width(), b.height()) * size;
		final Point3		cen = new Point3(b.right-dist, b.bottom-dist, 0);
		final double		scale = (90.0 / (double)(steps-1));

		ans.mPts.add(new Point3(b.left, b.top, 0));
		ans.mPts.add(new Point3(b.right, b.top, 0));
		
		for (int k=0; k<steps; k++) {
			final Point3	pt = new Point3(0, 0, 0);
			double			deg = (k*scale);
			getCircleEdge(cen, 360-deg, dist, pt);
			ans.mPts.add(pt);
		}
		
		ans.mPts.add(new Point3(b.left, b.bottom, 0));

		return ans;
	}

	// 0 = right, counterclockwise (90 = up)
	private final static void getCircleEdge(final Point3 cen, double degree, final double dist,
											final Point3 out) {
		degree *= Geo.DEGREES_TO_RADIANS;
		out.x = (cen.x + (Math.cos(degree) * dist));
		out.y = (cen.y - (Math.sin(degree) * dist));
	}

}
