package com.exh.bevel;

import java.util.LinkedList;
import java.util.List;

import android.graphics.Bitmap;
import android.graphics.Point;

import com.exh.bevel.algo.Clip;
import com.exh.bevel.algo.Extrude;
import com.exh.bevel.render.RenderTri;
import com.exh.bevel.render.Spotlight;
import com.exh.math.primitive.Geo;
import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Polygon;
import com.exh.math.primitive.Tri;

/* SHAPE-GENERATOR
 * The entry point for the algorithm.  Create the initial point list
 * input data, extrude it with the beveller, and convert it to a
 * bitmap with the renderer.
 */
public final class ShapeGenerator implements Generator {
	
	private final Point				mSize = new Point();
	
//	private final static float		MASTER_DEPTH = 428;
//	private final static float		MASTER_DEPTH = 225;
	
//	private final static float		MASTER_DEPTH = 195;
//	private final static float		MASTER_DEPTH = 125;
	private final static float		MASTER_DEPTH = 90;
//	private final static float		MASTER_DEPTH = 80;
//	private final static float		MASTER_DEPTH = 60;

//	private final static float		MASTER_DEPTH = 40;
//	private final static float		MASTER_DEPTH = 30;
	private final static double		BEVEL_AMOUNT = MASTER_DEPTH;
	
	private final static int		RENDER_F = (RenderTri.FILL | RenderTri.STROKE);
//	private final static int		RENDER_F = (Render.RENDER);
	
	// This determines the final Z value, which is scaled
	// according to the bevel amount.  Higher values mean
	// taller peaks, which means more pronounced lighting.
	private final static double		HEIGHT_SCALE = 0.75;
	
	/* SHAPE-GENERATOR
	 *****************/
	public ShapeGenerator() {
	}
	
	public Bitmap run(final ProgressReport progress) {
		// OK, I don't have to JUST create the silo shape.
//		final Input					pts = Input.initSquare();
//		final Input					pts = Input.initQuad();
//		final Input					pts = Input.initT();
		final Input					pts = Input.initThrowingStar();
//		final Input					pts = Input.initTree();
//		final Input					pts = Input.initCircle(5, 270);	// pentagon
//		final Input					pts = Input.initCircle(20, 0);
//		final Input					pts = Input.initCornerCircle(50, 0.25);
//		final Input					pts = Input.initSilo();
		
		final int					flags = RENDER_F | RenderTri.DEPTH | RenderTri.SPOTLIGHT;
//		final int					flags = RENDER_F;
		
		return run(pts, BEVEL_AMOUNT, flags, progress);
	}
	
	public Bitmap run(	final Input input, final double amount,
						final int flags, final ProgressReport progress) {
		progress.setProgressReport(0);
		
		translateToBounds(input.mPts, mSize);

		try {
			// On object used to highlight geometry, for error reporting,
			// and general poking into what's going on.
			final Spotlight					spot = new Spotlight();
			final double					height = amount * HEIGHT_SCALE;
			
			// Take the list of points and extrude them into 3d quads.
			final LinkedList<Tri>			result = new LinkedList<Tri>();
			final LinkedList<Polygon>		src = new LinkedList<Polygon>();
			final Extrude					ext = new Extrude();
			ext.on(input.mPts, input.mDirection, amount, src);
//	for (Point3 pt : input) System.out.println("SHAPE=" + pt);
//	for (Tri t : result) System.out.println("TRI=" + t);
			// Clip all the generated triangles by each other.

			final Clip						clipper = new Clip();
			clipper.mSpotlight = spot;
			clipper.mClipping = ((flags&RenderTri.NO_CLIP) == 0);
			clipper.on(new Point3(0, 0, amount*10), src, result, progress);
				
			// Eliminate any triangles outisde the original poly.
			scrubOverages(input.mPts, result);
				
			// Render the clipped quads.
			final RenderTri					render = new RenderTri();
			return render.newImage(result, mSize.x, mSize.y, flags, height, spot);
		} catch (Exception ex) {
			System.out.println("ERROR in root=" + ex);
		}
		return null;
	}

	// Given a list of points, make them fit inside the standard bounds.
	final static void translateToBounds(final List<Point3> pts, final Point outSize) {
		final float					BORDER = 4;
		final int					W = 256, H = 256;
		translateToBounds(pts, BORDER, BORDER, W-BORDER, H-BORDER);
		if (outSize != null) {
			outSize.x = W;
			outSize.y = H;
		}
	}

	// Given a list of points, make them fit inside the given bounds.
	final static void translateToBounds(final List<Point3> pts,
										final double left, final double top,
										final double right, final double bottom) {
		double			minX = pts.get(0).x, minY = pts.get(0).y;
		double			maxX = minX, maxY = minY;
		for (Point3 pt : pts) {
			if (pt.x < minX) minX = pt.x;
			else if (pt.x > maxX) maxX = pt.x;
			if (pt.y < minY) minY = pt.y;
			else if (pt.y > maxY) maxY = pt.y;
		}
		final double	scaleX = (right-left)/(maxX-minX),
						scaleY = (bottom-top)/(maxY-minY);
		for (Point3 pt : pts) {
			pt.x = left + ((pt.x-minX)*scaleX);
			pt.y = top + ((pt.y-minY)*scaleY);
		}
	}
	
	// Remove any tris that are external to the polygon by testing to 
	// see if their barycenters are outside the poly.
	private final void scrubOverages(final List<Point3> pts, final List<Tri> output) {
		final Polygon			p = new Polygon();
		final Point3			bary = new Point3();
		for (Point3 pt : pts) p.add(pt);
		p.add(p.mPt[0]);
		final LinkedList<Tri>	src = new LinkedList<Tri>();
		src.addAll(output);
		output.clear();
		
		for (Tri t : src) {
			t.getBarycenter(bary);
			if (Geo.polygonContains(p.mPt, p.mSize, bary.x, bary.y))
				output.add(t);
		}

	}

}