package com.exh.bevel.render;

import java.util.LinkedList;
import java.util.List;

import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.Path;

import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Tri;

public final class RenderTri {

	// Render fills every pixel and interpolates
	public final static int				RENDER = (1<<0);
	public final static int				SHADOWS = (1<<1);	// Ehh, ugly, don't use

	// Fill fills the quads with a solid colour, stroke draws their outlines
	public final static int				FILL = (1<<2);
	public final static int				STROKE = (1<<3);
	public final static int				DEPTH = (1<<6);

	// Spotlight highlights specific quads and geometry
	public final static int				SPOTLIGHT = (1<<4);

	// This doesn't belong in the render, it tells the bevel algorithm
	// to ignore clipping.  It's just hacked in for my screenshots
	public final static int				NO_CLIP = (1<<5);
	
	private final List<Surface>		mSurface = new LinkedList<Surface>();
	private final Normal				mNormal = new Normal();
	
	public final static int				FG_C = 0xff888888;
	public final static int				FILL_A = 160;
	public final static float			STROKE_THIN = 1.5f;
	public final static float			STROKE_THICK = 1.5f;
	
	private final static Point3			LIGHT = newLight(200, -100, 40);

	private final static Point3	newLight(final double x, final double y, final double z) {
		final Point3		pt = new Point3(x, y, z);
//		pt.normalize();
		return pt;
	}
	
	public Bitmap newImage(	final List<Tri> tl, final int w, final int h,
							final int flags, final double maxZ, final Spotlight spot) {
		/* SETUP MY SURFACES
		 */
System.out.println("RENDER size=" + tl.size());
		for (Tri t : tl) {
			final Surface	s = new Surface(t);
			mNormal.setOn(t, s);
//System.out.println("\tn=" + s.mNormal + " for " + t);
			mSurface.add(s);
		}
		if (mSurface.size() < 1) return null;
		// Cache the normal info from all adjacent surfaces.
//		fillAdjacentNormals();
		
		/* CREATE THE IMAGE AND DRAW
		 */
		final Bitmap		ans = Bitmap.createBitmap(w, h, Bitmap.Config.ARGB_8888);
		ans.eraseColor(0x00000000);
		if ((flags&RENDER) != 0) {
			System.out.println("RENDER OFF FOR NOW");
		}
		final Canvas		canvas = new Canvas(ans);
		final Paint			p = new Paint(Paint.ANTI_ALIAS_FLAG);
		if ((flags&FILL) != 0) {
			p.setStyle(Paint.Style.FILL);
			p.setStrokeWidth(1);
			onDraw(canvas, p, true);
		}
		if ((flags&STROKE) != 0) {
			p.setStyle(Paint.Style.STROKE);
			p.setColor(0x22000000);
			p.setStrokeWidth(STROKE_THIN);
			// A little darker if no fill
			if ((flags&FILL) == 0) {
				p.setColor(0xcc000000);
				p.setStrokeWidth(STROKE_THICK);
			}
			onDraw(canvas, p, false);
		}
		
		if ((flags&DEPTH) != 0) {
			onDepth(canvas, tl, maxZ);
		}
		
		if ((flags&SPOTLIGHT) != 0) {
			spot.onDraw(canvas);
		}
		
		{
			p.setStyle(Paint.Style.FILL);
			p.setStrokeWidth(1);
			p.setColor(0x8800ff00);
//			onDrawInteresting(canvas, p);
		}

		return ans;
	}
	
	private void onDepth(final Canvas canvas, final List<Tri> tl, final double maxZ) {
		final Canvas.VertexMode	mode = Canvas.VertexMode.TRIANGLES;
		final int				count = Tri.SIZE*2;
		final float[]			verts = new float[count];
		final int[]				colors = new int[count];	// Only uses half but still needs to be fully allocated
		final Paint				p = new Paint(Paint.ANTI_ALIAS_FLAG);
		final int				BLACK = Color.argb(128, 0, 0, 0);
		final double			LIMITER = 0.85;
		for (Tri t : tl) {
			for (int k=0; k<Tri.SIZE; k++) {
				final Point3	pt = t.mPt[k];
				verts[k*2] = (float)pt.x;
				verts[(k*2)+1] = (float)pt.y;

				if (pt.z <= 0) colors[k] = BLACK;
				else {
					int			c = (int)((pt.z/maxZ)*LIMITER*255);
					if (c < 0) c = 0; else if (c > 255) c = 255;
					colors[k] = Color.argb(128, c, c, c);
				}
			}
			
			canvas.drawVertices(mode, count, verts, 0, null, 0, colors, 0, null, 0, 0, p);
		}
	}
	
	private void onDraw(	final Canvas canvas, final Paint p,
							final boolean colorWithNormal) {
		final Path			path = new Path();
		int					c = -1;
System.out.println("onDraw size=" + mSurface.size());
		for (Surface e : mSurface) {
			c++;
//			if (c > 20 && c < 40) continue;
//			if (c == 5) continue;
//			if (c < 62 || c > 65) continue;
//			if (!(c < 62 || c > 65)) continue;
//System.out.println("\t" + c + " n=" + e.mNormal + " tri=" + e.mTri);
			path.rewind();
			Point3			lastPt = e.mTri.mPt[e.mTri.mPt.length-1];
			path.moveTo((float)lastPt.x, (float)lastPt.y);
			for (Point3 pt : e.mTri.mPt) {
				path.lineTo((float)pt.x, (float)pt.y);
			}
			path.close();
			if (colorWithNormal) {
				p.setColor(makeColor(FG_C, FILL_A, lightFrom(e.mNormal, LIGHT)));
			}
			canvas.drawPath(path, p);
		}
	}
	
	private void onDrawInteresting(final Canvas canvas, final Paint p) {
		final Path			path = new Path();
		int 				k = 0;
		for (Surface e : mSurface) {
			if (k >= 110 && k <= 110) {
				path.rewind();
				Point3			lastPt = e.mTri.mPt[e.mTri.mPt.length-1];
				path.moveTo((float)lastPt.x, (float)lastPt.y);
				for (Point3 pt : e.mTri.mPt) {
					path.lineTo((float)pt.x, (float)pt.y);
				}
				path.close();
				canvas.drawPath(path, p);
			}
			k++;
		}
	}

	private final double lightFrom(final Point3 normal, final Point3 light) {
		return (normal.x*light.x + normal.y*light.y + normal.z*light.z) / (normal.length() * light.length());
	}

	private final static int makeColor(final int srcC, final int srcA, final double scale) {
		int			r = Color.red(srcC), g = Color.green(srcC), b = Color.blue(srcC);
		r = (int)(r + (255*scale));
		if (r < 0) r = 0; else if (r > 255) r = 255;
		g = (int)(g + (255*scale));
		if (g < 0) g = 0; else if (g > 255) g = 255;
		b = (int)(b + (255*scale));
		if (b < 0) b = 0; else if (b > 255) b = 255;
		return Color.argb(srcA, r, g, b);
	}

	/* NORMAL
	 * Do the work of generating each surface normal
	 * The normal is the cross product of the vectors
	 * of two non-parallel sides of the surface.
	 ***********************************************/
	private final static class Normal {
		private final Point3			mA = new Point3(), mB = new Point3();
		
		// The surface normal of a quad is the cross product of its diagonals.
		// Realistically I could just take the normal of one triangle, which
		// is a little more efficient.
		void setOn(final Tri t, final Surface s) {
			mA.asVector(t.mPt[1], t.mPt[0]);
			mB.asVector(t.mPt[2], t.mPt[1]);
			s.mNormal.asCrossProduct(mA, mB);
			s.mNormal.normalize();
		}
	}
}