package com.exh.bevel.render;

import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Path;

import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Polygon;
import com.exh.math.primitive.Segment;
import com.exh.math.primitive.Tri;

/* A collection of annotations to the screen
 * to highlight features of interest.
 */
public final class Spotlight {
	public boolean					mActive = false;

	public final Polygon[]			mPoly = new Polygon[16];
	private final int[]				mPolyC = new int[mPoly.length];
	public final Point3[]			mPt = new Point3[16];
	public final Segment[]			mSeg = new Segment[16];
	public final Tri[]				mTri = new Tri[16];

	private final Paint				mP = new Paint(Paint.ANTI_ALIAS_FLAG);
	private final Path				mPath = new Path();
	
	public Spotlight() {
		for (int k=0; k<mPolyC.length; k++) mPolyC[k] = 0x770ff8ff;
	}
	
	public void clear() {
		for (int k=0; k<mPoly.length; k++) mPoly[k] = null;
		for (int k=0; k<mPt.length; k++) mPt[k] = null;
		for (int k=0; k<mSeg.length; k++) mSeg[k] = null;
		for (int k=0; k<mTri.length; k++) mTri[k] = null;
	}

	public void onDraw(final Canvas canvas) {		
		// Polygons
		for (int k=0; k<mPoly.length; k++) {
			final Polygon		poly = mPoly[k];
			if (poly == null) continue;
			mP.setColor(mPolyC[k]);
	System.out.println("DRAW ERR POLY=" + k); poly.print();
			onDrawPath(canvas, poly.mPt, poly.mSize);
		}
		
		// Segments
		mP.setStrokeWidth(2);
		for (int k=0; k<mSeg.length; k++) {
			final Segment		s = mSeg[k];
			if (s == null) continue;
System.out.println("DRAW ERR SEG " + k + " = " + s);
			mP.setColor(0xaa44ff66);
			canvas.drawLine((float)s.mPt[0].x, (float)s.mPt[0].y, (float)s.mPt[1].x, (float)s.mPt[1].y, mP);
		}
		mP.setStrokeWidth(1);

		for (Tri t : mTri) {
			if (t != null) {
				mP.setColor(0xaa00ff00);
				onDrawPath(canvas, t.mPt, Tri.SIZE);
			}
		}
		
		// Points
		for (int k=0; k<mPt.length; k++) {
			final Point3		pt = mPt[k];
			if (pt != null) {
System.out.println("DRAW ERR PT " + k + " = " + pt);
				mP.setColor(0x9952ff64);
				canvas.drawCircle((float)pt.x, (float)pt.y, 4, mP);
			}
		}
	}
	
	public void pushPoint(final Point3 pt) {
		for (int k=0; k<mPt.length; k++) {
			if (mPt[k] == null) {
				mPt[k] = new Point3(pt);
				return;
			}
		}
	}
	
	public void pushSegment(final Segment s) {
		for (int k=0; k<mSeg.length; k++) {
			if (mSeg[k] == null) {
				mSeg[k] = new Segment(s);
				return;
			}
		}
	}
	
	public void pushTri(final Tri t) {
		for (int k=0; k<mTri.length; k++) {
			if (mTri[k] == null) {
				mTri[k] = t.copy();
				return;
			}
		}
	}
	
	public void pushPoly(final Polygon p) {
		for (int k=0; k<mPoly.length; k++) {
			if (mPoly[k] == null) {
				mPoly[k] = p.copy();
				return;
			}
		}
	}

	private final void onDrawPath(	final Canvas canvas, final Point3[] pts,
									final int size) {
		mPath.rewind();
		Point3			pt = pts[size-1];
		mPath.moveTo((float)pt.x, (float)pt.y);
		for (int k=0; k<size; k++) {
			pt = pts[k];
			mPath.lineTo((float)pt.x, (float)pt.y);
		}
		canvas.drawPath(mPath, mP);
	}

}
