package com.exh.math.poly;

import java.util.Vector;

import com.exh.math.primitive.Geo;
import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Polygon;
import com.exh.math.tests.PolyIsExternal;
import com.exh.math.tests.PtInPoly;

/* A boolean operation on two polygons.  While this is mostly
 * general purpose code, it has been tweaked slightly to specifically
 * work with the one case we need it for, subtracting one triangle
 * from another.
 * 
 * Works by taking two lists of points and expanding them to include
 * all intersections with the other list, then selecting a start point
 * and walking the list, switching to the other at an intersection.
 * 
 * There are some aspects of the design that are a little weird because
 * it's intended eventually for C++.  So, for example, the intersection
 * points are stored in a separate list, which in C++ would be the object
 * responsible for managing the point memory.
 */
public final class Bool {

	
	final static boolean			TESTING_BUILD = false;
	final static boolean			TESTING_IO = false;
	
	// Return codes for on()
	public final static int			OK = 0;
	public final static int			ERROR = -1;
	
	// For now I'm just hard-limiting the number of possible answers
	// to 8, which covers what I need, but still.
	private final Results			mEdgeResults = new Results();
	
	private final StepList			mDst = new StepList(),
									mSrc = new StepList();
	private final EdgeList			mEdgeA = new EdgeList(),
									mEdgeB = new EdgeList();
	
	private final Flatten			mFlatten = new Flatten();
	private final BuildEdges		mBuildEdges = new BuildEdges();
	
	private final PolyIsExternal	mPiE = new PolyIsExternal();
	private final PtInPoly			mPP = new PtInPoly();
	
	private final BoolCutOut		mCutOut = new BoolCutOut();
	
	private final EvilDoerA			mEvilA = new EvilDoerA();
	
	private static int				COUNT = 0;
	
	public Bool() {
	}

	// If the result is valid, then swap the polygon data.
	public boolean takeResult(final int idx, final Polygon p) {
		return getResults().takeResult(idx, p);
	}

	// See the return codes defined above.
	public int subtract(	final Polygon dst, final Polygon src,
							final Vector<Point3> dp0, final Vector<Point3> dp1) {
COUNT++;
		final Results		r = getResults();
		r.start();
/*
System.out.println("***BOOL dst:");
dst.print();
System.out.println("Bool src:");
src.print();
*/
		if (!init(dst, src)) return ERROR;
		final int		rootCount = setAlignment(mDst, src, false);
		setAlignment(mSrc, dst, true);
/*
System.out.println("DST");
mDst.print();
System.out.println("SRC");
mSrc.print();
*/
		if (TESTING_IO) {
			System.out.println("\r\nBOOL subtract");
			System.out.println("roots=" + rootCount + " intersections=" + intersectionCount());
			System.out.println("ORIG DST");
			dst.print();
			System.out.println("DST");
			mDst.print();
			System.out.println("ORIG SRC");
			src.print();
			System.out.println("SRC");
			mSrc.print();
		}
//System.out.println("roots=" + rootCount);
		if (rootCount < 1) {
			if (TESTING_IO) System.out.println("BOOL RESULT root < 1, return empty");
			// If there are no roots, then dst is completely
			// covered by src, so we answer an empty set.
			// Technically, if we were actually operating on polygons,
			// this would not be true, but I have no clue how to
			// account for a case where a clip polygon clips out
			// every dst point.
			return OK;
		}
		final int		intersectionCount = intersectionCount();
//System.out.println("intersections=" + intersectionCount);
		if (intersectionCount < 1) {
			if (TESTING_IO) System.out.println("BOOL RESULT intersection < 1, error");
			// No intersections could mean one of two things:
			// Either there is no overlap, or src is completely
			// covered by dst.  We still need to implement that
			// second case by removing the "hole" dst creates.
			// If any point from src is inside dst, then dst completely
			// encompasses src, and a hole is created.
			final Point3		pt = src.mPt[0];
			if (Geo.polygonContains(dst.mPt, dst.mSize, pt.x, pt.y)) {
				r.mAns[0] = new Polygon();
				r.mAns[1] = new Polygon();
				if (mCutOut.on(dst, src, r.mAns[0], r.mAns[1])) {
					r.mSize = 2;
					return OK;
				}
			}
			return ERROR;
		} else if (intersectionCount == 1) {
			if (TESTING_IO) System.out.println("BOOL RESULT intersection = 1, return dst");
			// A single intersection means no clipping occurs -- the
			// clip poly doesn't penetrate the dst, it simply has
			// a colinear point.  The result is the input.
			r.mAns[0] = dst.copy();
			r.mSize = 1;
			return OK;
		}
		// We can hit a case where two polygons are aligned along
		// a single edge.  No clipping occurs.
//System.out.println("check external ct=" + COUNT);
//			if (COUNT < -216) {
//			if (COUNT == 216) {
			if (COUNT > -216) {
		if (mPiE.on(dst, src)) {
// System.out.println("EXTERNAL!");
//			System.out.println("PiE c=" + COUNT);
//			System.out.println("pie dst"); dst.print();
//			System.out.println("pie src"); src.print();
			if (TESTING_IO) System.out.println("BOOL RESULT clip is external, return dst");
			r.mAns[0] = dst.copy();
			r.mSize = 1;
			return OK;
			}
		}

		mEvilA.mPoly = src;
		// Prepare the edges, which guide the building process
		if (!mEdgeA.start(mDst, mSrc, true, mEvilA, "D")) return ERROR;
		if (!mEdgeB.start(mSrc, mDst, false, null, "S")) return ERROR;
		mBuildEdges.on(mEdgeA, mEdgeB, r);
		return OK;
	}

	private boolean init(final Polygon dst, final Polygon src) {
		try {
			mFlatten.start();
			mFlatten.setFindIntersectionSource(src, null);
			mDst.flatten(dst, mFlatten);
			mFlatten.setFindIntersectionSource(null, mDst);
			mSrc.flatten(src, mFlatten);
			return true;
		} catch (Exception ex) {
			System.out.println("ERROR init()=" + ex);
		}
		return false;
	}

	private int setAlignment(final StepList a, final Polygon b, final boolean insideIsGood) {
		int					goodCount = 0;
		for (int k=0; k<a.mSize; k++) {
			final Step		s = a.mStep[k];
			if (s.isIntersection()) continue;

			final boolean	inside = mPP.on(s.mPt, b);
			if (inside) {
				if (insideIsGood) s.mAlignment = Step.GOOD;
				else s.mAlignment = Step.EVIL;
			} else {
				if (insideIsGood) s.mAlignment = Step.EVIL;
				else s.mAlignment = Step.GOOD;
			}
			if (s.mAlignment == Step.GOOD) goodCount++;
		}
		return goodCount;
	}
	
	private int intersectionCount() {
		int			ans = 0;
		for (int k=0; k<mDst.mSize; k++) {
			if (mDst.mStep[k].mIntersection != null) ans++;
		}
		return ans;
	}

	// If the result is valid, then swap the polygon data.
	public Results getResults() {
		return mEdgeResults;
	}
	
	/* RESULTS
	 * A place to store polygon lists, mostly useful since I'm experimenting
	 * with different building algorithms.
	 */
	final static class Results {
		Polygon[]				mAns = new Polygon[16];		// REALLY unlikely I'd produce more than this
		int						mSize;
		
		void start() {
			mSize = 0;
		}
		
		boolean add(final Polygon p) {
			if (mSize >= mAns.length) return false;
			if (Bool.TESTING_IO) {
				System.out.println("output " + mSize);
				p.print();
			}
			mAns[mSize] = p;
			mSize++;
			return true;
		}
		
		public boolean takeResult(final int idx, final Polygon p) {
			if (idx < 0 || idx >= mSize) return false;
			p.clear();
			if (mAns[idx] != null) p.swap(mAns[idx]);
			return true;
		}
	}
	
	/* EVIL-DOER-A
	 * Determine whether or not an edge on A is EVIL.
	 */
	private final static class EvilDoerA implements EvilDoer {

		Polygon						mPoly;
		
		private final PtInPoly		mPP = new PtInPoly();
		private final Point3		mPt = new Point3();
		
		public void on(final Edge e) {
			if (e.mAlignment <= Step.EVIL) return;
			// Take the midpoint of the edge - if it's inside the
			// poly, then the edge is invalid.
			mPt.asMid(e.mA.mPt, e.mB.mPt);
			if (mPP.on(mPt, mPoly)) e.mAlignment = Step.EVIL;
		}
	}

}
