package com.exh.math.poly;

import com.exh.math.primitive.Polygon;

/* Convert two EdgeLists into a new polygon by finding
 * the next step during the build process.
 */
class BuildEdges {
	
	// Idx is always the idx into mA.
	private Source			mA = new Source(),
							mB = new Source();

	void on(final EdgeList a, final EdgeList b, final Bool.Results r) {
		r.start();
		
		Polygon				p;
		while ((p = nextPoly(a, b)) != null) {
			if (p.mSize > 2) r.add(p);
			if (r.mSize >= r.mAns.length) return;
		}
	}

	private Polygon nextPoly(final EdgeList a, final EdgeList b) {
		// Find the next available starting point.  To be a starting
		// point, you need an edge that's not bad, with an a pt that's good.
		int					aIdx = -1;
		for (int k=0; k<a.mSize; k++) {
			final Edge		e = a.mEdge[k];
			if (e.mAlignment >= Step.NEUTRAL && e.mA.mAlignment >= Step.GOOD) {
				aIdx = k;
				break;
			}
		}
		if (aIdx < 0) return null;
		// Generate a polygon starting on the a pt.  It can't be
		// any larger then the size of both lists, so use that as
		// an error condition.  Otherwise stop when we hit a cycle.
		int					errorSize = a.mSize + b.mSize + 1;
		final Edge			cycle = a.mEdge[aIdx];
		Edge				e;
		final Polygon		p = new Polygon();
		final Step			cycleStep = cycle.mA;
		start(aIdx, a, b);
		p.add(cycle.mA.mPt);
		if (Bool.TESTING_BUILD) System.out.println("Build pt " + cycle.mAName);
		cycle.mA.mAlignment = Step.EVIL;
		while ((e = nextEdge()) != null) {
			if (cycleStep == e.mB) return p;
			// Indicates we ran out of unused points, not an error condition
			if (e.mAlignment == Step.EVIL) return p;
			if (errorSize-- <= 0) {
				if (Bool.TESTING_BUILD) System.out.println("ERROR Bool.nextPoly() hit a cycle");
				return null;
			}
			
			// Mark this edge as used
//			if (e.mAlignment == Step.GOOD) e.mAlignment = Step.EVIL;
			if (e.mAlignment >= Step.GOOD) e.mAlignment = Step.EVIL;
			p.add(e.mB.mPt);
			if (Bool.TESTING_BUILD) System.out.println("Build pt " + e.mBName);
		}
		if (Bool.TESTING_BUILD) System.out.println("ERROR Bool.nextPoly() answered a null step");
		return null;
	}

	// idxA must be a valid index into a mStep.
	void start(final int idxA, final EdgeList a, final EdgeList b) {
		if (Bool.TESTING_BUILD) System.out.println("START BUILD idx=" + idxA);
		mA.start(idxA, a, "D");
		mB.start(0, b, "S");
	}

	Edge nextEdge() {
		final Edge			nxt = mA.cur();
		mA.step();
//		if (Bool.TESTING_BUILD) System.out.println("\tnextEdge=" + mA.db());
		// If the endpoint is an intersection, need to decide where to go next.
		// By default I should switch to b, but if that's bad, stick with a.
		if (nxt.mB.isIntersection()) {
			if (!mB.seek(nxt.mB.mIntersection)) return null;
			if (mB.cur().mAlignment > Step.EVIL) {
				switchSources();
				if (Bool.TESTING_BUILD) System.out.println("\t\tintersection, swap, new=" + mA.db());
			}
		}
	//	if (Bool.TESTING_BUILD) System.out.println("\treturning edgealign=" + nxt.mAlignment);
		return nxt;
	}

	private void switchSources() {
		final Source		s = mA;		mA = mB;	mB = s;
	}
	
	// SOURCE class
	// Just a place to store the current step list info.
	private final static class Source {
		EdgeList		mEdge;
		// The current index.  Only valid for the "a" source, except in
		// cases where we're peeking ahead.
		int				mIdx;
		String			mPrefix;
		
		void start(final int idx, final EdgeList a, final String prefix) {
			mEdge = a;
			mIdx = idx;
			mPrefix = prefix;
		}
		
		String db() {
			return mPrefix + mIdx;
		}
		
		Edge cur() {
			return mEdge.mEdge[mIdx];
		}
		
		// Step my list and answer with the current alignment
		void step() {
			mIdx += 1;
			if (mIdx >= mEdge.mSize) mIdx = 0;
		}
		
		boolean seek(final Intersection i) {
			mIdx = mEdge.findIndexForA(i);
			if (mIdx < 0 || mIdx >= mEdge.mSize) {
				System.out.println("ERROR Build.seek() intersection missing");
				return false;
			}
			return true;
		}
	}
}
