package com.exh.math.poly;

import com.exh.math.primitive.Point3;

/* A single step in a flattened point list.
 */
final class Step {

	// I don't own the point.  It is either owned by the
	// source polygon or the intersection.
	Point3				mPt;
	
	// Each step is characterized as GOOD, NEUTRAL or EVIL.  Neutral
	// is synonymous for being an intersection; Good indicates this point
	// must be in the final output, Evil indicates it can't be.  Good points
	// can only be used once, so once they are, they become Evil.
	static final int	GOOD = 1;
	static final int	NEUTRAL = 0;
	static final int	EVIL = -1;
	int					mAlignment;
	
	// My intersection, if any.  I don't own this object.
	Intersection		mIntersection;
	
	void set(final Point3 pt) {
		mPt = pt;
		mAlignment = GOOD;
		mIntersection = null;
	}
	
	void set(final Intersection inter) {
		mPt = inter.mPt;
		mAlignment = NEUTRAL;
		mIntersection = inter;
	}
	
	boolean isIntersection() {
		return mIntersection != null;
	}
	
	public String toString() {
		final String	s = mPt.toString() + " intersection=" + (mIntersection != null) + " (" + mIntersection + ")";
		if (mAlignment == GOOD) return s + " GOOD";
		else if (mAlignment == EVIL) return s + " EVIL";
		else if (mAlignment != NEUTRAL) return s + " ? " + mAlignment;
		return s;
	}

}
