package com.exh.math.poly;

import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Polygon;

/* A list of Steps, flattened from an incoming polygon.
 */
final class StepList {

	Step[]				mStep = null;
	int					mSize = 0;

	StepList() {
		growTo(16);
	}
	
	// Add a step for each point in p, as well as any
	// intersections with interP.  If List inter is supplied,
	// then it will contain all intersections.
	void flatten(final Polygon p, final Flatten flatten) {
//System.out.println("FLATTEN");
		mSize = 0;
		if (p.mSize < 2) return;
		
		Point3			prv = p.mPt[p.mSize-1];
		for (int k=0; k<p.mSize; k++) {
			final Point3	nxt = p.mPt[k];
			flatten.addIntersections(prv, nxt, this);
			add(nxt, flatten.mEps);
			prv = nxt;
		}
		// If the last and first steps are the same, remove the
		// last, but maintain any intersection.
		if (sameAsLast(mStep[0].mPt, flatten.mEps)) {
			if (mStep[mSize-1].isIntersection()) mStep[0].mIntersection = mStep[mSize-1].mIntersection;
			mSize--;
		}
	}
	
	// Add the point, provided it's unique from the previous.
	void add(final Point3 pt, final double eps) {
		if (mSize >= mStep.length) growTo(mStep.length+8);

		if (sameAsLast(pt, eps)) return;
		
		mStep[mSize].set(pt);
		mSize++;
	}

	// Add the point, provided it's unique from the
	// previous.  If it's not, then convert the existing
	// point into an intersection.
	void add(final Intersection inter, final double eps) {
		if (mSize >= mStep.length) growTo(mStep.length+8);

		if (sameAsLast(inter.mPt, eps)) {
			mStep[mSize-1].set(inter);
		} else {
			mStep[mSize].set(inter);
			mSize++;
		}
	}

	// Answer true if pt is within eps distance of my tail point, if any.
	private boolean sameAsLast(final Point3 pt, final double eps) {
		if (mSize <= 0) return false;
		return pt.sameAs(mStep[mSize-1].mPt, eps);
	}
	
	int findIndexFor(final Intersection inter) {
		for (int k=0; k<mSize; k++) {
			final Step		s = mStep[k];
			if (s.mIntersection == inter) return k;
		}
		return -1;
	}

	private void growTo(final int newLength) {
		if (mStep == null || newLength > mStep.length) {
			final Step[]		newStep = new Step[newLength];
			final int			oldLength = (mStep != null ? mStep.length : 0);
			if (oldLength > 0) System.arraycopy(mStep, 0, newStep, 0, mStep.length);
			for (int k=oldLength; k<newLength; k++) newStep[k] = new Step();
			mStep = newStep;
		}
	}
	
	void print() {
		for (int k=0; k<mSize; k++) {
			System.out.println("\t" + k + " " + mStep[k].toString());
		}
	}
}
