package com.exh.math.primitive;

import android.graphics.Path;
import android.graphics.RectF;

/* I always store my own point data.
 */
public class Polygon {

	public Point3[]			mPt = null;
	public int				mSize = 0;

	public void clear() {
		mSize = 0;
	}
	
	public Polygon copy() {
		final Polygon		ans = new Polygon();
		if (mSize < 1) return ans;
		ans.mPt = new Point3[mSize];
		ans.mSize = mSize;
		for (int k=0; k<mSize; k++) ans.mPt[k] = new Point3(mPt[k].x, mPt[k].y, mPt[k].z);
		return ans;
	}
	
	public void swap(final Polygon o) {
		final Point3[]		pt = mPt;	mPt = o.mPt;		o.mPt = pt;
		final int			s = mSize;	mSize = o.mSize;	o.mSize = s;
	}
	
	public void setAlloc(final int newAlloc) {
		if (newAlloc <= 0) {
			mSize = 0;
			return;
		}
		if (mPt == null || mPt.length < newAlloc) {
			final Point3[]		newPt = new Point3[newAlloc];
			if (mPt != null) System.arraycopy(mPt, 0, newPt, 0, mPt.length);
			mPt = newPt;
		}
	}
	
	public void getFrame(final RectF f) {
		if (mSize < 1) {
			f.set(0, 0, -1, -1);
			return;
		}
		f.left = (float)mPt[0].x;
		f.top = (float)mPt[0].y;
		f.right = f.left;
		f.bottom = f.top;
		for (int k=1; k<mSize; k++) {
			final Point3		pt = mPt[k];
			f.left = Math.min(f.left, (float)pt.x);
			f.top = Math.min(f.top, (float)pt.y);
			f.right = Math.max(f.right, (float)pt.x);
			f.bottom = Math.max(f.bottom, (float)pt.y);
		}
	}
	
	public void getBarycenter(final Point3 bary) {
		Geo.getBarycenter(mPt, mSize, bary);
	}
	
	public void rewind() {
		mSize = 0;
	}
	
	public Point3 add() {
		if (mPt == null || mSize >= mPt.length) setAlloc(mSize+4);
		
		if (mPt[mSize] == null) mPt[mSize] = new Point3();
		mSize++;
		return mPt[mSize-1];
	}
	
	public void add(final Point3 pt) {
		add().set(pt);
	}
	
	public void add(final double x, final double y, final double z) {
		add().set(x, y, z);
	}
	
	public void remove(final int idx) {
		if (idx < 0 || idx >= mSize) return;

		for (int k=idx; k<mSize-1; k++) {
			mPt[k].set(mPt[k+1]);
		}
		mSize--;
	}
	
	// Remove all adjacent points within the given tolerance, i.e.
	// remove what we have loosely identified as "duplicates."
	public void compress(final double eps) {
		for (int k=1; k<mSize; k++) {
			while (mSize > k && mPt[k-1].sameAs(mPt[k], eps)) {
				for (int j=k+1; j<mSize; j++) {
					mPt[j-1].set(mPt[j]);
				}
				mSize--;
			}
		}
		// Check first - last
		if (mSize > 2 && mPt[0].sameAs(mPt[mSize-1], eps)) mSize--;
	}

	// Answer true if pt is within eps distance of my tail point, if any.
	public boolean sameAsLast(final Point3 pt, final double eps) {
		if (mSize <= 0) return false;
		return pt.sameAs(mPt[mSize-1], eps);
	}
	
	public void setTo(final Point3[] pts) {
		setTo(pts, pts.length);
	}
	
	public void setTo(final Point3[] pts, final int size) {
		mSize = 0;
		for (int k=0; k<size; k++) add(pts[k]);
	}
	
	public void setTo(final Polygon p) {
		setTo(p.mPt, p.mSize);
	}

	public void setTo(final Tri t) {
		mSize = 0;
		for (Point3 pt : t.mPt) add(pt);
	}

	public boolean includes(final Point3 pt, final double eps) {
		return includesSquared(pt, eps*eps);
	}

	public boolean includesSquared(final Point3 pt, final double eps) {
		for (int k=0; k<mSize; k++) {
			if (mPt[k].dist3Squared(pt) <= eps) return true;
		}
		return false;
	}
	
	// Drawing/testing
	public void addToPath(final Path p) {
		if (mSize < 2) return;
		
		Point3			pt = mPt[mSize-1];
		p.moveTo((float)pt.x, (float)pt.y);
		for (int k=0; k<mSize; k++) {
			pt = mPt[k];
			p.lineTo((float)pt.x, (float)pt.y);
		}
	}

	public String toString() {
		return "Polygon size=" + mSize;
	}

	public void print() {
		System.out.println("Polygon size=" + mSize);
		for (int k=0; k<mSize; k++) System.out.println("\t" + k + "=" + mPt[k]);
	}

}
