package com.exh.math.tests;

import com.exh.math.primitive.Point3;

/* Distance between two lines (not segments), 3d
 * From http://www.mathworks.com/matlabcentral/newsreader/view_thread/246420
 */
public class LineLine {

public boolean  mPrint = false;
	// After evaluating, the nearest points on line 1 and line 2.
	public final Point3		mNearest1 = new Point3(),
							mNearest2 = new Point3();

	private final Point3	mB2B1 = new Point3(),
							mA1B1 = new Point3(),
							mA2A1 = new Point3(),
							mU = new Point3(),
							mV = new Point3();

	public LineLine() {
	}

	public double	on(	final Point3 a1, final Point3 a2,
						final Point3 b1, final Point3 b2) {
		mB2B1.asSub(b2, b1);
		mA1B1.asSub(a1, b1);
		mA2A1.asSub(a2, a1);
	
		mU.asCrossProduct(mB2B1, mA1B1);
		mV.asCrossProduct(mA2A1, mB2B1);
		final double		nA = mU.dot3(mV);
		mU.asCrossProduct(mA2A1, mA1B1);
		mV.asCrossProduct(mA2A1, mB2B1);
		final double		nB = mU.dot3(mV);
		mU.asCrossProduct(mA2A1, mB2B1);
		mV.asCrossProduct(mA2A1, mB2B1);
		final double		t = mU.dot3(mV);
	
		mA2A1.mult(nA/t);
		mNearest1.asAdd(a1, mA2A1);
		mB2B1.mult(nB/t);
		mNearest2.asAdd(b1, mB2B1);
if (mPrint) System.out.println("\t\tn1=" + mNearest1 + " 2=" + mNearest2);
if (mNearest1.sameAs(mNearest2, 0.00000001)) {
	if (mPrint) System.out.println("\t\tANS EPS");
			return 0.00000001;
}
		return mNearest1.dist3(mNearest2);
	}
	
	
	public double			mEps = 0.00001;
	public final Point3 	mNearestA = new Point3(),
							mNearestB = new Point3();

	private double			mMua, mMub;
	
	private final Point3	m13 = new Point3(),
							m43 = new Point3(),
							m21 = new Point3();
	
	public boolean	on2(final Point3 a1, final Point3 a2,
						final Point3 b1, final Point3 b2) {
		m13.asSub(a1, b1);
		m43.asSub(b2, b1);
if (mPrint) System.out.println("\tLL on2() a 13 " + m13 + " 43 " + m43);
		if (Math.abs(m43.x) < mEps && Math.abs(m43.y) < mEps && Math.abs(m43.z) < mEps)
			return false;
if (mPrint) System.out.println("\tLL on2() b");
		m21.asSub(a2, a1);
		if (Math.abs(m21.x) < mEps && Math.abs(m21.y) < mEps && Math.abs(m21.z) < mEps)
			return false;
if (mPrint) System.out.println("\tLL on2() c");
	
		final double	d1343 = m13.x * m43.x + m13.y * m43.y + m13.z * m43.z;
		final double	d4321 = m43.x * m21.x + m43.y * m21.y + m43.z * m21.z;
		final double	d1321 = m13.x * m21.x + m13.y * m21.y + m13.z * m21.z;
		final double	d4343 = m43.x * m43.x + m43.y * m43.y + m43.z * m43.z;
		final double	d2121 = m21.x * m21.x + m21.y * m21.y + m21.z * m21.z;
	
		final double	denom = d2121 * d4343 - d4321 * d4321;
		if (Math.abs(denom) < mEps)
			return false;
		final double	numer = d1343 * d4321 - d1321 * d4343;
if (mPrint) System.out.println("\tLL on2() d");
	
		mMua = numer / denom;
		mMub = (d1343 + d4321 * (mMua)) / d4343;
	
		mNearestA.x = a1.x + mMua * m21.x;
		mNearestA.y = a1.y + mMua * m21.y;
		mNearestA.z = a1.z + mMua * m21.z;
		mNearestB.x = b1.x + mMub * m43.x;
		mNearestB.y = b1.y + mMub * m43.y;
		mNearestB.z = b1.z + mMub * m43.z;
	
	   return true;
	}
}



/* MIGHT BE MORE EFFICIENT
 * 
http://local.wasp.uwa.edu.au/~pbourke/geometry/lineline3d/

C source

typedef struct {
   double x,y,z;
} XYZ;


   Calculate the line segment PaPb that is the shortest route between
   two lines P1P2 and P3P4. Calculate also the values of mua and mub where
      Pa = P1 + mua (P2 - P1)
      Pb = P3 + mub (P4 - P3)
   Return FALSE if no solution exists.

int LineLineIntersect(
   XYZ p1,XYZ p2,XYZ p3,XYZ p4,XYZ *pa,XYZ *pb,
   double *mua, double *mub)
{
   XYZ p13,p43,p21;
   double d1343,d4321,d1321,d4343,d2121;
   double numer,denom;

   p13.x = p1.x - p3.x;
   p13.y = p1.y - p3.y;
   p13.z = p1.z - p3.z;
   p43.x = p4.x - p3.x;
   p43.y = p4.y - p3.y;
   p43.z = p4.z - p3.z;
   if (ABS(p43.x)  < EPS && ABS(p43.y)  < EPS && ABS(p43.z)  < EPS)
      return(FALSE);
   p21.x = p2.x - p1.x;
   p21.y = p2.y - p1.y;
   p21.z = p2.z - p1.z;
   if (ABS(p21.x)  < EPS && ABS(p21.y)  < EPS && ABS(p21.z)  < EPS)
      return(FALSE);

   d1343 = p13.x * p43.x + p13.y * p43.y + p13.z * p43.z;
   d4321 = p43.x * p21.x + p43.y * p21.y + p43.z * p21.z;
   d1321 = p13.x * p21.x + p13.y * p21.y + p13.z * p21.z;
   d4343 = p43.x * p43.x + p43.y * p43.y + p43.z * p43.z;
   d2121 = p21.x * p21.x + p21.y * p21.y + p21.z * p21.z;

   denom = d2121 * d4343 - d4321 * d4321;
   if (ABS(denom) < EPS)
      return(FALSE);
   numer = d1343 * d4321 - d1321 * d4343;

   *mua = numer / denom;
   *mub = (d1343 + d4321 * (*mua)) / d4343;

   pa->x = p1.x + *mua * p21.x;
   pa->y = p1.y + *mua * p21.y;
   pa->z = p1.z + *mua * p21.z;
   pb->x = p3.x + *mub * p43.x;
   pb->y = p3.y + *mub * p43.y;
   pb->z = p3.z + *mub * p43.z;

   return(TRUE);
}
*/