package com.exh.math.tests;

import com.exh.math.primitive.Plane;
import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Segment;

// Intersection between a segment and a plane
// http://softsurfer.com/Archive/algorithm_0104/algorithm_0104B.htm
public class SegPlane {

	public static final int		INTERSECT = 1;		// Success, outIp is valid
	public static final int		INSIDE = 0;			// Success? Depends on the context: segment lies inside plane
	public static final int		ERROR = -1;
	public static final int		DISJOINT = ERROR;
	
	private static final double	EPS = 0.000001;
	
	private final Point3		mU = new Point3(), mW = new Point3();
	
	// Segment L is actually an unbounded line, not a segment
	public int					on(Segment S, Plane Pn, Point3 outIp) {
		mU.sub(S.mPt[1], S.mPt[0]);
		mW.sub(S.mPt[0], Pn.mPt);
		
		final double		D = Pn.mN.dot3(mU),
							N = -Pn.mN.dot3(mW);
		
		if (Math.abs(D) < EPS) {			// segment is parallel to plane
			if (N == 0) return INSIDE;		// segment lies in plane
			else return DISJOINT;			// no intersection                 
		}
		
		// they are not parallel
		// compute intersect param
		final double		sI = N / D;
		if (sI < 0 || sI > 1) return DISJOINT;	// no intersection
		
		// compute segment intersect point
		if (outIp != null) {
			mU.mult(sI);
			outIp.add(S.mPt[0], mU);
//			*I = S.P0 + sI * u;
		}
		return INTERSECT;
	}
}
